from typing import Any, Iterator, List

from ..data.base_reader import BaseReader
from ..data_generators.base_data_generator import BaseDataGenerator
from ..schemas.common_structures import InputData
from ..schemas.data_generator_configs import BaseDataGeneratorConfig
from ..schemas.dataset_config import DatasetConfig, DatasetSourceType
from ..schemas.reader_configs import BaseReaderConfig


class DataProcessor:
    """
    Utility class to process data based on DatasetConfig.
    """

    def __init__(self, config: Any):
        self.config = DatasetConfig(**config)

    def process_data(self) -> Iterator[List[InputData]]:
        """
        Processes data based on the DatasetConfig and returns the processed
        data.
        """

        # If the source type is a dataset, use a reader to process it
        if self.config.source_type == DatasetSourceType.DATASET.value:
            return self._process_dataset()

        # If the source type is machine-generated, use data generators
        elif self.config.source_type == DatasetSourceType.MACHINE_GENERATED.value:
            return self._process_machine_generated()

        else:
            raise ValueError(f"Unknown source type: {self.config.source_type}")

    def _process_dataset(self) -> Iterator[List[InputData]]:
        """
        Processes data from a dataset.
        """

        if not self.config.reader or not self.config.file_path:
            return iter([])  # return empty iterator

        # Instantiate the reader based on the config's reader attribute
        if self.config.reader:

            reader_cls = BaseReader.get_reader(self.config.reader)

            if reader_cls:
                config_cls = BaseReader.get_config_class(self.config.reader)
                if config_cls:
                    if self.config.reader_config:
                        if isinstance(self.config.reader_config, dict):
                            config_data = self.config.reader_config
                        else:
                            config_data = self.config.reader_config.asdict()
                        config_instance = config_cls(**config_data)
                    else:
                        config_instance = config_cls()
                    reader = reader_cls(config_instance)
                    return reader.read(self.config.file_path)
                else:
                    reader = reader_cls(BaseReaderConfig())
                    return reader.read(self.config.file_path)
        return iter([])

    def _process_machine_generated(self) -> Iterator[List[InputData]]:
        """
        Processes machine-generated data.
        """
        if self.config.data_generators:
            for data_generator, data_generator_config in self.config.data_generators.items(
            ):
                data_generator_cls = BaseDataGenerator.get_data_generator(
                    data_generator
                )
                if data_generator_cls:
                    config_cls = BaseDataGenerator.get_config_class(
                        data_generator
                    )
                    if config_cls:
                        if isinstance(data_generator_config, dict):
                            config_data = data_generator_config
                        else:
                            config_data = data_generator_config.asdict()
                        config_instance = config_cls(**config_data)
                        data_generator_instance = data_generator_cls(
                            config_instance
                        )
                        return data_generator_instance.generate_examples()
                    else:
                        data_generator_instance = data_generator_cls(
                            BaseDataGeneratorConfig()
                        )

        # Implement logic for processing machine-generated data
        return iter([])



import io
from contextlib import redirect_stdout
from functools import wraps

from .base_parser import BaseParserWithRegistry


def capture_and_parse_with_base_registry(config=None):
    """
    Decorator to capture stdout of a function and parse it using a specified
    parser.

    The parser is determined based on the provided configuration. If the
    specified parser is not found in the BaseParserWithRegistry's registry,
    the function's output will not be captured or parsed.

    Args:
    - config (dict, optional): Configuration dict with 'parser' key specifying
                               the parser class name.

    Returns:
    - Decorated function that captures and parses its stdout.
    """
    parser_name = config.get("parser", None) if config else None
    parser_instance = None

    if parser_name and parser_name in BaseParserWithRegistry.registry:
        parser_class = BaseParserWithRegistry.registry[parser_name]
        parser_instance = parser_class()

    def decorator(func):

        @wraps(func)
        def wrapper(*args, **kwargs):
            if parser_instance:
                f = io.StringIO()
                with redirect_stdout(f):
                    original_output = func(*args, **kwargs)
                captured_output = f.getvalue()
                return original_output, parser_instance.parse(captured_output)
            else:
                original_output = func(*args, **kwargs)
                return original_output, []

        return wrapper

    return decorator



import base64
import io
import json
from typing import List

import pandas as pd  # type: ignore
from dash import html  # type: ignore
from PIL import Image

from yival.schemas.experiment_config import GroupedExperimentResult


def sanitize_group_key(group_key: str):
    try:
        json_str = group_key.replace('example_id:', '"example_id":').replace(
            'content:', '"content":'
        ).replace('expected_result:', '"expected_result":')
        escaped_json_str = json_str.replace("\n", "\\n").replace("\t", "\\t")
        valid_json_str = escaped_json_str.replace(": None", ": null")
        data_dict = json.loads(valid_json_str)
        content = data_dict['content']
        if content:
            items = [str(value) for value in content.values()]
            group_key = ", ".join([item.strip() for item in items])
            return group_key
    except Exception:
        return group_key
    return ""


def sanitize_column_name(name):
    return name.replace('"', '').replace(':', '')


import pandas as pd


def highlight_best_values(df: pd.DataFrame, *cols) -> list:

    def numeric_col_best_value_style(col: str) -> dict:
        """Return style for best value in a numeric column."""
        best_val = df[col].min()
        return {
            'if': {
                'filter_query': f'{{{col}}} eq {best_val}',
                'column_id': col
            },
            'border': '2px solid red'
        }

    def non_numeric_col_best_value_style(col: str) -> dict:
        """Return style for best value in a non-numeric column containing
           metrics."""
        metrics_values = df[col].str.extractall(r":\s?(\d+\.\d+)"
                                                ).astype(float)
        if not metrics_values.empty:
            best_val = metrics_values[0].min()
            return {
                'if': {
                    'filter_query': f'{{{col}}} contains "{best_val}"',
                    'column_id': col
                },
                'border': '2px solid red'
            }
        return {}

    styles = []
    for col in df.columns:
        if pd.api.types.is_numeric_dtype(df[col]):
            styles.append(numeric_col_best_value_style(col))
        else:
            style = non_numeric_col_best_value_style(col)
            if style:  # Check if the style was generated for the non-numeric column
                styles.append(style)

    return styles


def image_to_base64(image: Image.Image) -> str:
    '''Converts an image to base64 string.'''
    buffered = io.BytesIO()
    image.save(buffered, format="PNG")
    img_str = base64.b64encode(buffered.getvalue()).decode()
    return img_str


def process_raw_output(raw_output):
    if isinstance(raw_output, list) and all(
        isinstance(item, Image.Image) for item in raw_output
    ):
        return [image_to_base64(image) for image in raw_output]
    else:
        return raw_output


def generate_group_key_combination_data(
    group_experiment_results: List[GroupedExperimentResult]
) -> pd.DataFrame:
    data_list = []
    all_combos = set()
    for group in group_experiment_results:
        group_key = group.group_key
        group_key = sanitize_group_key(group_key)
        if group_key == "":
            pass
        row_dict = {"Test Data": group_key}
        for exp_result in group.experiment_results:
            combo_str = sanitize_column_name(
                str(exp_result.combination).replace("{", "").replace("}", "")
            )
            nested_output = {
                "text_output":
                process_raw_output(exp_result.raw_output.text_output),
                "image_output":
                process_raw_output(
                    getattr(exp_result.raw_output, 'image_output', None)
                ),
                "video_output":
                process_raw_output(
                    getattr(exp_result.raw_output, 'video_output', None)
                ),
                "evaluator_outputs":
                "\n".join([
                    f"{e.name} : {e.display_name} = {e.result}"
                    for e in exp_result.evaluator_outputs
                ]) if exp_result.evaluator_outputs else None
            }
            video_text = ''
            video_output = nested_output['video_output']
            if video_output is not None:
                for video_url in nested_output['video_output']:
                    video_text += f"<yival_video_output>{video_url}</yival_video_output>"

            formatted_output = f"<yival_raw_output>{nested_output['text_output']}</yival_raw_output>{nested_output['image_output']}{video_text}\n{nested_output['evaluator_outputs']}"
            row_dict[combo_str] = formatted_output
            all_combos.add(combo_str)
        data_list.append(row_dict)

    df = pd.DataFrame(data_list)
    # Ensure all combo columns exist
    for combo in all_combos:
        if combo not in df.columns:
            df[combo] = None

    return df


def generate_heatmap_style(df, *cols):
    styles = []

    light_positive = (173, 216, 230)
    dark_positive = (70, 130, 180)

    light_negative = (255, 245, 235)
    dark_negative = (255, 150, 140)

    for col in df.columns:
        if col == "Iteration":
            continue
        if pd.api.types.is_numeric_dtype(df[col]):
            min_val = df[col].min()
            max_val = df[col].max()
            range_val = max_val - min_val if min_val != max_val else 1
            for val in df[col].unique():
                normalized = (val - min_val) / range_val

                if col in ["Average Token Usage", "Average Latency"]:
                    bg_color = tuple(
                        int(
                            light_negative[i] + normalized *
                            (dark_negative[i] - light_negative[i])
                        ) for i in range(3)
                    )
                else:
                    bg_color = tuple(
                        int(
                            light_positive[i] + normalized *
                            (dark_positive[i] - light_positive[i])
                        ) for i in range(3)
                    )

                bg_color_str = f"rgb({bg_color[0]}, {bg_color[1]}, {bg_color[2]})"
                styles.append({
                    'if': {
                        'filter_query': f'{{{col}}} eq {val}',
                        'column_id': col
                    },
                    'backgroundColor': bg_color_str
                })
        else:  # For aggregated metrics columns and evaluator outputs
            metrics_values = df[col].str.extractall(r":\s?(\d+\.\d+)"
                                                    ).astype(float)
            if not metrics_values.empty:
                min_val = metrics_values[0].min()
                max_val = metrics_values[0].max()
                range_val = max_val - min_val if min_val != max_val else 1
                for _, val in metrics_values[0].items():
                    normalized = (val - min_val) / range_val
                    bg_color = tuple(
                        int(
                            light_positive[i] + normalized *
                            (dark_positive[i] - light_positive[i])
                        ) for i in range(3)
                    )

                    bg_color_str = f"rgb({bg_color[0]}, {bg_color[1]}, {bg_color[2]})"
                    styles.append({
                        'if': {
                            'filter_query': f'{{{col}}} contains "{val}"',
                            'column_id': col
                        },
                        'backgroundColor': bg_color_str
                    })
    return styles


def generate_legend():
    """
    Generates the legend for the heatmap.
    """
    # Define the colors for the legend
    colors = [("Low Value (Positive)", "rgb(173, 216, 230)"),
              ("High Value (Positive)", "rgb(70, 130, 180)"),
              ("Low Value (Negative)", "rgb(255, 245, 235)"),
              ("High Value (Negative)", "rgb(255, 150, 140)")]

    # Create the legend elements
    legend_elements = [
        html.Div([
            html.Div(
                style={
                    "backgroundColor": color,
                    "width": "20px",
                    "height": "20px",
                    "display": "inline-block"
                }
            ),
            html.Span(f" {label}", style={"marginLeft": "10px"})
        ]) for label, color in colors
    ]

    # Return the legend wrapped in a container
    return html.Div(
        legend_elements, style={
            "marginTop": "20px",
            "marginBottom": "20px"
        }
    )



# type: ignore

import os
import pickle

from yival.experiment.utils import get_function_args
from yival.schemas.experiment_config import Experiment


def interactive_bot(
    experiment_data: Experiment,
    experiment_config,
    all_combinations,
    state,
    logger,
    evaluator,
    interactive=True,
):
    print("Running interactive bot mode...")
    if experiment_data.enhancer_output:
        for group_result in experiment_data.enhancer_output.group_experiment_results:
            experiment_results = []
            seen = set()
            for r in group_result.experiment_results:
                if str(r.combination) in seen:
                    continue
                else:
                    seen.add(str(r.combination))
                    experiment_results.append(r)
            group_result.experiment_results = experiment_results
    function_args = get_function_args(
        experiment_config["custom_function"], experiment_config["dataset"]
    )
    function_args["yival_expected_result (Optional)"] = 'str'

    data = {
        "experiment_data": experiment_data,
        "experiment_config": experiment_config,
        "function_args": function_args,
        "all_combinations": all_combinations,
        "state": state,
        "logger": logger,
        "evaluator": evaluator,
        "interactive": interactive,
    }
    with open("data.pkl", "wb") as f:
        pickle.dump(data, f)

    os.system("streamlit run src/yival/experiment/bot/run_streamlit.py")


if __name__ == "__main__":
    pass



"""
BERTScore is a language model evaluation metric based on the BERT language
model.
It leverages the pre-trained contextual embeddings from BERT and matches words
in candidate and reference sentences by cosine similarity. 
It has been shown to correlate with human judgment on sentence-level and
system-level evaluation.
"""

from bert_score import score  # type:ignore

from ..schemas.common_structures import InputData
from ..schemas.evaluator_config import BertScoreEvaluatorConfig, EvaluatorOutput
from ..schemas.experiment_config import ExperimentResult, MultimodalOutput
from .base_evaluator import BaseEvaluator


class BertScoreEvaluator(BaseEvaluator):
    """Evaluator calculate bert_score"""
    default_config = BertScoreEvaluatorConfig(name="bertsocre_evaluator")

    def __init__(self, config: BertScoreEvaluatorConfig):
        super().__init__(config)
        self.config = config

    def evaluate(self, experiment_result: ExperimentResult) -> EvaluatorOutput:
        """Evaluate the experiment result according to bertsocre"""
        assert isinstance(self.config, BertScoreEvaluatorConfig)
        input_data = experiment_result.input_data
        raw_output = experiment_result.raw_output.text_output
        expected_result = input_data.expected_result

        p, r, f1 = score([raw_output], [expected_result],
                         lang=self.config.lan,
                         verbose=True)

        scores = {"p": p.item(), "r": r.item(), "f": f1.item()}
        result = scores.get(self.config.indicator, .0)

        return EvaluatorOutput(
            name=self.config.name,
            display_name=self.config.display_name,
            result=result,
            metric_calculators=self.config.metric_calculators
        )


BaseEvaluator.register_evaluator(
    "bertscore_evaluator", BertScoreEvaluator, BertScoreEvaluatorConfig
)


def main():
    """Main function to test the bertscore evaluator"""
    evaluator_config = BertScoreEvaluatorConfig(
        name="bertscore_evaluator",
        display_name="rouge",
        metric_calculators=[],
    )
    input_data_example = InputData(
        content={
            "instruction": "translate the sentence to english",
        },
        expected_result="Have a great day!"
    )
    experiment_result_example = ExperimentResult(
        input_data=input_data_example,
        combination={
            "wrapper1": "var1",
            "wrapper2": "var2"
        },
        raw_output=MultimodalOutput(text_output="have a nice day"),
        latency=30.0,
        token_usage=20
    )

    evaluator = BertScoreEvaluator(evaluator_config)
    result = evaluator.evaluate(experiment_result_example)
    print(result)


if __name__ == "__main__":
    main()


import numpy as np
import openai


def get_embedding(input_str: str) -> list[float]:
    result = openai.Embedding.create(
        model='text-embedding-ada-002',
        input=input_str,
    )
    return result['data'][0]['embedding']


def cosine_similarity(a: list[float], b: list[float]) -> float:
    return np.dot(a, b) / (np.linalg.norm(a) * np.linalg.norm(b))


def get_cosine_simarity(doc1: str, doc2: str) -> float:
    doc1_embedding = get_embedding(doc1)
    doc2_embedding = get_embedding(doc2)
    return cosine_similarity(doc1_embedding, doc2_embedding)



import asyncio
import importlib
import inspect
import json
import os
import sys
import time
from collections import defaultdict
from importlib import import_module
from typing import Any, Dict, List

from yival.variation_generators.chain_of_density_prompt import ChainOfDensityPromptGenerator

from ..data.base_reader import BaseReader
from ..data.csv_reader import CSVReader
from ..data.huggingface_dataset_reader import HuggingFaceDatasetReader
from ..data_generators.base_data_generator import BaseDataGenerator
from ..enhancers.base_combination_enhancer import BaseCombinationEnhancer
from ..evaluators.alpaca_eval_evaluator import AlpacaEvalEvaluator
from ..evaluators.base_evaluator import BaseEvaluator
from ..evaluators.bertscore_evaluator import BertScoreEvaluator
from ..evaluators.openai_elo_evaluator import OpenAIEloEvaluator
from ..evaluators.openai_prompt_based_evaluator import OpenAIPromptBasedEvaluator
from ..evaluators.python_validation_evaluator import PythonValidationEvaluator
from ..evaluators.rouge_evaluator import RougeEvaluator
from ..evaluators.string_expected_result_evaluator import StringExpectedResultEvaluator
from ..finetune.base_trainer import BaseTrainer
from ..logger.token_logger import TokenLogger
from ..result_selectors.ahp_selection import AHPSelection
from ..result_selectors.selection_strategy import SelectionStrategy
from ..schemas.combination_enhancer_configs import BaseCombinationEnhancerConfig
from ..schemas.evaluator_config import MethodCalculationMethod
from ..schemas.experiment_config import (
    CombinationAggregatedMetrics,
    Experiment,
    ExperimentConfig,
    ExperimentResult,
    GroupedExperimentResult,
    InputData,
    Metric,
)
from ..schemas.selector_strategies import BaseConfig
from ..schemas.trainer_configs import BaseTrainerConfig
from ..states.experiment_state import ExperimentState
from ..variation_generators.base_variation_generator import BaseVariationGenerator
from ..variation_generators.openai_prompt_based_variation_generator import (
    OpenAIPromptBasedVariationGenerator,
)
from ..wrappers.base_wrapper import BaseWrapper
from ..wrappers.string_wrapper import StringWrapper
from .evaluator import Evaluator


def is_async_function(func):
    return asyncio.iscoroutinefunction(func)


def import_function_from_string(func_string: str):
    """Helper function to import a function from a string."""
    try:
        # Direct import for built-in functions
        module_name, function_name = func_string.rsplit('.', 1)
        module = importlib.import_module(module_name)
    except ImportError:
        # Dynamic import for user-specified functions
        module_path, module_name, function_name = func_string.rsplit('.', 2)
        sys.path.append(os.path.abspath(module_path))
        module = importlib.import_module(module_name)

    function = getattr(module, function_name)
    return function


def get_yaml_args(dataset):
    """Get argument from yaml."""
    return dataset['data_generators']['openai_prompt_data_generator'][
        'input_function']['parameters']


def get_function_args(func_string: str, dataset: dict):
    """Get argument types of a function."""
    if func_string == "demo.complete_task.complete_task":
        args = get_yaml_args(dataset)
        return {name: f"<class '{param}'>" for name, param in args.items()}
    else:
        function = import_function_from_string(func_string)
        signature = inspect.signature(function)

        return {
            name: str(param.annotation)
            for name, param in signature.parameters.items()
            if name.lower() != "state"
        }


def call_function_from_string(func_string: str, **kwargs) -> Any:
    """Call a function specified by a string."""
    function = import_function_from_string(func_string)
    return function(**kwargs)


async def acall_function_from_string(func_string: str, **kwargs) -> Any:

    function = import_function_from_string(func_string)
    if is_async_function(function):
        return await function(**kwargs)
    else:
        return function(**kwargs)


def _add_to_sys_path_if_not_present(directory: str):
    """Add directory to sys.path if not already present."""
    if directory not in sys.path:
        sys.path.append(directory)


def _get_class_from_path(cls_path: str):
    """Retrieve a class from a module using its full path."""
    module_path, class_name = cls_path.rsplit(".", 1)
    directory, base_module_name = os.path.split(module_path)

    _add_to_sys_path_if_not_present(directory)
    return getattr(import_module(base_module_name), class_name)


def register_custom_readers(custom_readers: Dict[str, Dict[str, Any]]):
    for name, details in custom_readers.items():
        reader_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseReader.register_reader(name, reader_cls, config_cls)

    _ = CSVReader
    _ = HuggingFaceDatasetReader


def register_custom_enhancer(custom_enhancer: Dict[str, Dict[str, Any]]):
    for name, details in custom_enhancer.items():
        enhancer_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseCombinationEnhancer.register_enhancer(
            name, enhancer_cls, config_cls
        )
    from ..enhancers.openai_prompt_based_combination_enhancer import (
        OpenAIPromptBasedCombinationEnhancer,
    )
    _ = OpenAIPromptBasedCombinationEnhancer


def register_custom_selection_strategy(
    custom_strategy: Dict[str, Dict[str, Any]]
):
    for name, details in custom_strategy.items():
        strategy_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        SelectionStrategy.register_strategy(name, strategy_cls, config_cls)
    _ = AHPSelection


def register_custom_evaluators(custom_evaulators: Dict[str, Dict[str, Any]]):
    for name, details in custom_evaulators.items():
        evaluator_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseEvaluator.register_evaluator(name, evaluator_cls, config_cls)
    _ = StringExpectedResultEvaluator
    _ = OpenAIEloEvaluator
    _ = PythonValidationEvaluator
    _ = AlpacaEvalEvaluator
    _ = BertScoreEvaluator
    _ = OpenAIPromptBasedEvaluator
    _ = RougeEvaluator


def register_custom_wrappers(custom_wrappers: Dict[str, Dict[str, Any]]):
    for name, details in custom_wrappers.items():
        wrapper_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseWrapper.register_wrapper(name, wrapper_cls, config_cls)

    _ = StringWrapper


def register_custom_data_generator(
    custom_data_generators: Dict[str, Dict[str, Any]]
):
    for name, details in custom_data_generators.items():
        data_generator_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseDataGenerator.register_data_generator(
            name, data_generator_cls, config_cls
        )
    from ..data_generators.openai_prompt_data_generator import OpenAIPromptDataGenerator
    _ = OpenAIPromptDataGenerator


def register_custom_variation_generators(
    custom_variation_generators: Dict[str, Dict[str, Any]]
):
    for name, details in custom_variation_generators.items():
        variation_generator_cls = _get_class_from_path(details["class"])

        config_cls = None
        if "config_cls" in details:
            config_cls = _get_class_from_path(details["config_cls"])

        BaseVariationGenerator.register_variation_generator(
            name, variation_generator_cls, config_cls
        )
    _ = OpenAIPromptBasedVariationGenerator
    _ = ChainOfDensityPromptGenerator


def calculate_metrics(
    results: List[ExperimentResult]
) -> Dict[str, List[Metric]]:
    if not results:
        return {}

    res: Dict[str, List[Metric]] = defaultdict(list)

    # Assuming the metrics method across all results are the same,
    # so we'll use results[0] just to get a reference to the metric_calculators
    reference_evaluator_outputs = results[0].evaluator_outputs
    if reference_evaluator_outputs:
        for evaluator_output in reference_evaluator_outputs:
            for metric_calculator in evaluator_output.metric_calculators:
                if metric_calculator[
                    "method"
                ] == MethodCalculationMethod.AVERAGE.value:  # type: ignore
                    # Calculate average across all results
                    total = sum(
                        eo.result for r in results if r.evaluator_outputs
                        for eo in r.evaluator_outputs
                        if eo.name == evaluator_output.name
                        and eo.display_name == evaluator_output.display_name
                    )
                    average_value = total / len(results)
                    key = evaluator_output.name
                    key += ": " + evaluator_output.display_name if evaluator_output.display_name else ""
                    res[key].append(
                        Metric(
                            name=MethodCalculationMethod.AVERAGE.value,
                            value=average_value
                        )
                    )

    return res


def calculate_average_token(results: List[ExperimentResult]) -> float:
    if not results:
        return 0

    total = sum(r.token_usage for r in results)
    return total / len(results)


def calculate_average_latency(results: List[ExperimentResult]) -> float:
    if not results:
        return 0

    total = sum(r.latency for r in results)
    return total / len(results)


def run_single_input(
    d: InputData, config: ExperimentConfig, all_combinations: List[Dict[str,
                                                                        Any]],
    logger: TokenLogger, evaluator: Evaluator
):
    results = []
    tmp_state = ExperimentState()
    tmp_state.active = True
    for combo in all_combinations:
        for name, variation in combo.items():
            tmp_state.set_specific_variation(name, variation)
        start_time = time.time()
        res = call_function_from_string(
            config["custom_function"],  # type: ignore
            **d.content,
            state=tmp_state
        ) if "custom_function" in config else None  #type: ignore

        end_time = time.time()
        latency = end_time - start_time  # Time in seconds

        tokens_used = logger.get_current_usage()
        result = ExperimentResult(
            input_data=d,
            combination=combo,
            raw_output=res,
            latency=latency,
            token_usage=tokens_used,
            evaluator_outputs=[]
        )
        if result.evaluator_outputs is not None:
            result.evaluator_outputs.extend(
                asyncio.run(evaluator.evaluate_individual_result(result))
            )
        results.append(result)
    return results


async def arun_single_input(
    d: InputData, config: ExperimentConfig, all_combinations: List[Dict[str,
                                                                        Any]],
    logger: TokenLogger, evaluator: Evaluator
):
    results = []
    tmp_state = ExperimentState()
    tmp_state.active = True
    for combo in all_combinations:
        for name, variation in combo.items():
            tmp_state.set_specific_variation(name, variation)
        start_time = time.time()
        res = await acall_function_from_string(
            config["custom_function"],  # type: ignore
            **d.content,
            state=tmp_state
        )
        end_time = time.time()
        latency = end_time - start_time  # Time in seconds

        tokens_used = logger.get_current_usage()
        result = ExperimentResult(
            input_data=d,
            combination=combo,
            raw_output=res,
            latency=latency,
            token_usage=tokens_used,
            evaluator_outputs=[]
        )
        if result.evaluator_outputs is not None:
            result.evaluator_outputs.extend(
                asyncio.run(evaluator.evaluate_individual_result(result))
            )
        results.append(result)
    return results


def get_selection_strategy(
    config: ExperimentConfig
) -> SelectionStrategy | None:
    if "selection_strategy" not in config:  # type: ignore
        return None
    if config["selection_strategy"]:  # type: ignore
        for strategy, strategy_config in config["selection_strategy"
                                                ].items(  # type: ignore
                                                ):
            strategy_cls = SelectionStrategy.get_strategy(strategy)
            if strategy_cls:
                config_cls = SelectionStrategy.get_config_class(strategy)
                if config_cls:
                    if isinstance(strategy_config, dict):
                        config_data = strategy_config
                    else:
                        config_data = strategy_config.asdict()
                    config_instance = config_cls(**config_data)
                    strategy_instance = strategy_cls(config_instance)
                else:
                    strategy_instance = strategy_cls(BaseConfig())

                return strategy_instance
    return None


def get_enhancer(config: ExperimentConfig) -> BaseCombinationEnhancer | None:
    if "enhancer" not in config:  # type: ignore
        return None
    if config["enhancer"]:  # type: ignore
        enhancer_config = config["enhancer"]  # type: ignore
        enhancer_cls = BaseCombinationEnhancer.get_enhancer(
            enhancer_config["name"]
        )
        if enhancer_cls:
            config_cls = BaseCombinationEnhancer.get_config_class(
                enhancer_config["name"]
            )
            if config_cls:
                if isinstance(enhancer_config, dict):
                    config_data = enhancer_config
                else:
                    config_data = enhancer_config.asdict()
                config_instance = config_cls(**config_data)
                enhancer_instance = enhancer_cls(config_instance)
                return enhancer_instance
            else:
                enhancer_instance = enhancer_cls(
                    BaseCombinationEnhancerConfig(name="")
                )
                return enhancer_instance
    return None


def remove_none_values(d):
    if isinstance(d, dict):
        return {
            k: remove_none_values(v)
            for k, v in d.items() if v is not None
        }
    else:
        return d


def get_trainer(config: ExperimentConfig) -> BaseTrainer | None:
    if "trainer" not in config:  #type: ignore
        return None
    if config["trainer"]:  #type: ignore
        trainer_config = config["trainer"]  #type: ignore
        trainer_cls = BaseTrainer.get_trainer(trainer_config["name"])
        if trainer_cls:
            config_cls = BaseTrainer.get_config_class(trainer_config["name"])
            default_config = BaseTrainer.get_default_config(
                trainer_config["name"]
            )
            if config_cls:
                if isinstance(trainer_config, dict):
                    config_data = trainer_config
                else:
                    config_data = trainer_config.asdict()
                #remove None value
                config_data = remove_none_values(config_data)
                #update data use default_config
                default_config = default_config.asdict()  #type: ignore
                default_config.update(config_data)  #type: ignore
                config_instance = config_cls(**default_config)  #type: ignore
                trainer_instance = trainer_cls(config_instance)  #type: ignore
                return trainer_instance
            else:
                trainer_instance = trainer_cls(BaseTrainerConfig(name=""))
                return trainer_instance
    return None


def generate_experiment(
    results: List[ExperimentResult],
    evaluator: Evaluator,
    evaluate_all: bool = True,
    evaluate_group: bool = True
) -> Experiment:
    grouped_experiment_results: List[GroupedExperimentResult] = defaultdict(
        list
    )  # type: ignore

    for item in results:
        key = str(item.input_data)
        grouped_experiment_results[key].append(item)  # type: ignore
    grouped_experiment_results = [
        GroupedExperimentResult(group_key=k, experiment_results=v)
        for k, v in grouped_experiment_results.items()  # type: ignore
    ]
    if evaluate_group:
        for grouped_experiment_result in grouped_experiment_results:
            evaluator.evaluate_group_result(
                grouped_experiment_result.experiment_results
            )

    enable_custom_func = False
    if results[0].raw_output:
        enable_custom_func = True

    combo_metrics = defaultdict(list)
    for item in results:
        combo_str = json.dumps(item.combination)
        combo_metrics[combo_str].append(item)

    cobo_aggregated_metrics = [
        CombinationAggregatedMetrics(
            combo_key=k, experiment_results=v, aggregated_metrics={}
        ) for k, v in combo_metrics.items()
    ]

    for cobo_aggregated_metric in cobo_aggregated_metrics:
        cobo_aggregated_metric.aggregated_metrics = calculate_metrics(
            cobo_aggregated_metric.experiment_results
        )
        cobo_aggregated_metric.average_token_usage = calculate_average_token(
            cobo_aggregated_metric.experiment_results
        )
        cobo_aggregated_metric.average_latency = calculate_average_latency(
            cobo_aggregated_metric.experiment_results
        )

    experiment = Experiment(
        enable_custom_func=enable_custom_func,
        group_experiment_results=grouped_experiment_results,
        combination_aggregated_metrics=cobo_aggregated_metrics
    )

    if evaluate_all:
        er = [experiment]
        evaluator.evaluate_based_on_all_results(er)
    return experiment



"""
Variation Generator Base Module.

This module defines the `BaseVariationGenerator` class, an abstract base class
for all variation generators. A variation generator is responsible for
producing a sequence of variations to be used in experiments. The variations
could represent different configurations, parameter settings, or other factors
that should be tested in the context of an experiment. Subclasses of this base
class should implement the specific logic for generating the desired
variations.
"""

from abc import ABC, abstractmethod
from typing import Any, Dict, Iterator, List, Optional, Type

from ..schemas.experiment_config import WrapperVariation
from ..schemas.varation_generator_configs import BaseVariationGeneratorConfig


class BaseVariationGenerator(ABC):
    """
    Abstract base class for all variation variation.
    """
    _registry: Dict[str, Dict[str, Any]] = {}
    default_config: Optional[BaseVariationGeneratorConfig] = None

    @classmethod
    def get_variation_generator(
        cls, name: str
    ) -> Optional[Type['BaseVariationGenerator']]:
        """Retrieve variation generator class from registry by its name."""
        generator_info = cls._registry.get(name, {})
        return generator_info.get(
            "class", None
        ) if "class" in generator_info else None

    @classmethod
    def get_default_config(
        cls, name: str
    ) -> Optional[BaseVariationGeneratorConfig]:
        """Retrieve the default configuration of a variation generator by its
        name."""
        generator_info = cls._registry.get(name, {})
        return generator_info.get(
            "default_config", None
        ) if "default_config" in generator_info else None

    def __init__(self, config: BaseVariationGeneratorConfig):
        self.config = config

    @classmethod
    def get_config_class(
        cls, name: str
    ) -> Optional[Type[BaseVariationGeneratorConfig]]:
        """Retrieve the configuration class of a generator_info by its name."""
        generator_info = cls._registry.get(name, {})
        return generator_info.get("config_cls", None)

    @classmethod
    def register_variation_generator(
        cls,
        name: str,
        variation_generator_cls: Type['BaseVariationGenerator'],
        config_cls: Optional[Type[BaseVariationGeneratorConfig]] = None
    ):
        cls._registry[name] = {
            "class": variation_generator_cls,
            "default_config": variation_generator_cls.default_config,
            "config_cls": config_cls
        }

    @abstractmethod
    def generate_variations(self) -> Iterator[List[WrapperVariation]]:
        """
        Generate a sequence of variations to be used in experiments.
        
        This method should yield lists of variations, with each list typically
        representing a set or batch of variations to be used in a single
        experiment or iteration.
        
        Returns:
            Iterator[List[WrapperVariation]]: An iterator yielding lists of
            WrapperVariation objects.
            
        Note:
            The specific logic for generating variations should be implemented
            by subclasses.
        """


"""
Demo code to translate text from English to Chinese using GPT-3.
"""
import os

import openai

from yival.logger.token_logger import TokenLogger
from yival.schemas.experiment_config import MultimodalOutput
from yival.states.experiment_state import ExperimentState
from yival.wrappers.string_wrapper import StringWrapper


def translate(input: str, state: ExperimentState) -> MultimodalOutput:
    """
    Demo code to translate text from English to Chinese using GPT-3.
    """
    logger = TokenLogger()
    logger.reset()
    # Ensure you have your OpenAI API key set up
    openai.api_key = os.getenv("OPENAI_API_KEY")

    # Create a chat message sequence
    messages = [{
        "role":
        "system",
        "content":
        "You are a helpful assistant that translates English to any language."
    }, {
        "role":
        "user",
        "content":
        str(
            StringWrapper(
                "Translate the following to Chinese",
                name="translate",
                state=state
            )
        ) + f'{input}'
    }]
    response = openai.ChatCompletion.create(
        model="gpt-3.5-turbo", messages=messages
    )

    # Extract the assistant's message (translated text) from the response
    translated_text = MultimodalOutput(
        text_output=response['choices'][0]['message']['content'],
    )
    token_usage = response['usage']['total_tokens']
    logger.log(token_usage)

    return translated_text



"""
Wrapper Base Module.

This module introduces the `BaseWrapper` class, which serves as the
fundamental structure for wrappers in the experimental framework. Wrappers
manage experiment variations based on the global experiment state. They are
crucial components to control and monitor different variations during an
experiment's lifecycle.
"""

from typing import Any, Dict, Optional, Type

from ..schemas.wrapper_configs import BaseWrapperConfig
from ..states.experiment_state import ExperimentState


class BaseWrapper:
    """
    Base class for wrappers that manage experiment variations based on the
    global experiment state.
    
    This class provides the fundamental structure and methods for wrappers.
    Specific wrappers should inherit from this class and implement the
    necessary methods.
    
    Attributes:
        name (str): The name of the wrapper.
        experiment_state (ExperimentState): The global state of the experiment.
        config (BaseWrapperConfig): The configuration for the wrapper.
    """
    _registry: Dict[str, Dict[str, Any]] = {}
    default_config: Optional[BaseWrapperConfig] = None

    @classmethod
    def decorator_register(cls, name: str):
        """Decorator to register new wrappers."""

        def inner(subclass: Type[BaseWrapper]):
            cls._registry[name] = {
                "class": subclass,
                "default_config": subclass.default_config
            }
            return subclass

        return inner

    def __init__(
        self,
        name: str,
        config: Optional[BaseWrapperConfig] = None,
        state: Optional[ExperimentState] = None,
    ) -> None:
        self.name = name
        self.experiment_state = state if state else ExperimentState()
        self.config = config

    def get_variation(self) -> Optional[Any]:
        """
        Retrieve the next variation for the experiment based on the wrapper's
        name.

        Returns:
            Optional[Any]: The next variation if the experiment is active,
            otherwise None.
        """
        if self.experiment_state.active:
            return self.experiment_state.get_next_variation(self.name)
        return None

    @classmethod
    def get_wrapper(cls, name: str) -> Optional[Type['BaseWrapper']]:
        return cls._registry.get(name, {}).get("class")

    @classmethod
    def get_default_config(cls, name: str) -> Optional[BaseWrapperConfig]:
        return cls._registry.get(name, {}).get("default_config")

    @classmethod
    def get_config_class(cls, name: str) -> Optional[Type[BaseWrapperConfig]]:
        return cls._registry.get(name, {}).get("config_cls")

    def get_active_config(self, name: str) -> Optional[BaseWrapperConfig]:
        if self.experiment_state.active and self.experiment_state.config and self.experiment_state.config.wrapper_configs:
            config = self.experiment_state.config.wrapper_configs.get(name)
            if config:
                config_cls = BaseWrapper.get_config_class(name)
                if config_cls:
                    return config_cls(**config.asdict())
        return None

    @classmethod
    def register_wrapper(
        cls,
        name: str,
        wrapper_cls: Type['BaseWrapper'],
        config_cls: Optional[Type[BaseWrapperConfig]] = None
    ):
        cls._registry[name] = {
            "class": wrapper_cls,
            "default_config": wrapper_cls.default_config,
            "config_cls": config_cls
        }



from dataclasses import dataclass
from typing import Dict, List, Optional


# Configurations
@dataclass
class BaseConfig:
    pass  # Common configuration attributes for all selection methods


@dataclass
class AHPConfig(BaseConfig):
    criteria: List[str]
    criteria_weights: Dict[str, float]
    criteria_maximization: Dict[str, bool]
    normalize_func: Optional[str] = None


@dataclass
class SelectionOutput:
    best_combination: str
    selection_reason: Optional[Dict[str, float]] = None



from typing import cast

from omegaconf import OmegaConf
from omegaconf.listconfig import ListConfig

from ..schemas.experiment_config import ExperimentConfig


def load_and_validate_config(config_filepath: str) -> ExperimentConfig:
    """
    Load and validate the experiment configuration directly from a specified
    filepath.
    """

    # Load the configuration directly from the specified filepath
    dict_config = OmegaConf.load(config_filepath)

    # Convert the DictConfig to ExperimentConfig and cast it
    config = cast(ExperimentConfig, OmegaConf.to_object(dict_config))

    return config


def load_and_validate_configs(config_filepath: str) -> list[ExperimentConfig]:
    """
    Load and validate the experiment configurations
    """

    dict_configs = OmegaConf.load(config_filepath)
    if isinstance(dict_configs, ListConfig):
        configs = [
            cast(ExperimentConfig, OmegaConf.to_object(dc))
            for dc in dict_configs
        ]
    else:
        configs = [cast(ExperimentConfig, OmegaConf.to_object(dict_configs))]

    return configs
